#include "ltr030api.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>



typedef enum LTR030_USB_CMD {
    LTR030_USB_CMD_WRITE_DSP_FLASH =   9,
    LTR030_USB_CMD_READ_DSP_FLASH  =  10
} t_ltr030_usb_cmds;

#define LTR030_USB_MAX_CONTROL_LEN      4096

#define LTR030_FLASH_ADDR_BOOTLOADER    0x000000UL
#define LTR030_FLASH_ADDR_SERIAL        0x020000UL
#define LTR030_FLASH_ADDR_DSP_FIRM      0x020200UL
#define LTR030_FLASH_ADDR_CONFIG        0x1E6E00UL
#define LTR030_FLASH_ADDR_FPGA_INFO     0x1E7000UL
#define LTR030_FLASH_ADDR_FPGA_FIRM     0x1E7008UL

#define LTR030_FLASH_SIZE               0x200000UL

#define LTR030_FLASH_FPGA_FRIM_MAX_SIZE  (LTR030_FLASH_SIZE - LTR030_FLASH_ADDR_FPGA_FIRM)
#define LTR030_FLASH_DSP_FIRM_MAX_SIZE  (LTR030_FLASH_ADDR_CONFIG-LTR030_FLASH_ADDR_DSP_FIRM)

#define LTR030_FPGA_INFO_SIZE          8
#define LTR030_CONFIG_SIZE             14



#define LTR030_CONFIG_INTF_USB          0
#define LTR030_CONFIG_INTF_ETH          1


static const TLTR_ERROR_STRING_DEF f_err_str_tbl[] = {
    {LTR030_ERR_UNSUPORTED_CRATE_TYPE, "Данный тип крейта не поддерживается библиотекой"},
    {LTR030_ERR_FIRM_VERIFY,           "Ошибка проверки правильности записи прошивки"}
};


//-----------------------------------------------------------------------------
LTR030API_DllExport(INT) LTR030_Init(TLTR030 *module) {
    INT err =  module == NULL ? LTR_ERROR_INVALID_MODULE_DESCR : LTR_OK;
    if (err==LTR_OK) {
        memset(module, 0, sizeof(TLTR030));
        module->Size = sizeof(TLTR030);
        err = LTR_Init(&module->Channel);
    }

    return err;
}

//-----------------------------------------------------------------------------
LTR030API_DllExport(INT) LTR030_Open(TLTR030 *module, DWORD saddr, WORD sport,
                                     WORD iface, const CHAR *csn) {
    INT res = module == NULL ? LTR_ERROR_INVALID_MODULE_DESCR : LTR_OK;
    if (res==LTR_OK)
        res = LTR030_Close(module);
    if (res == LTR_OK) {
        module->Channel.saddr = saddr;
        module->Channel.sport = sport;

        if (csn!=NULL) {
            strncpy(module->Channel.csn, csn, LTR_CRATE_SERIAL_SIZE);
            module->Channel.csn[LTR_CRATE_SERIAL_SIZE-1] = '\0';
        } else {
            memset(module->Channel.csn, '\0', LTR_CRATE_SERIAL_SIZE);
        }

        module->Channel.cc = LTR_CC_CHNUM_CONTROL | ((iface & 0x7) << 8);

        res = LTR_Open(&module->Channel);
        if (res==LTR_OK)
            res = LTR_GetCrateInfo(&module->Channel, &module->Info);
        if ((res==LTR_OK) && (module->Info.CrateType!=LTR_CRATE_TYPE_LTR030)
                && (module->Info.CrateType!=LTR_CRATE_TYPE_LTR031)) {
            res = LTR030_ERR_UNSUPORTED_CRATE_TYPE;
            LTR030_Close(module);
        }
    }
    return res;
}
//-----------------------------------------------------------------------------
LTR030API_DllExport(INT)  LTR030_Close(TLTR030 *module) {
    return module == NULL ? LTR_ERROR_INVALID_MODULE_DESCR : LTR_Close(&module->Channel);
}

LTR030API_DllExport(INT) LTR030_FlashWrite(TLTR030 *module, DWORD addr, const BYTE* data, DWORD size) {
    INT err = module == NULL ? LTR_ERROR_INVALID_MODULE_DESCR : data == NULL ? LTR_ERROR_PARAMETERS :
                                                                               LTR_OK;
    while ((err==LTR_OK) && size) {
        DWORD snd_size = size;
        if (snd_size > LTR030_USB_MAX_CONTROL_LEN)
            snd_size = LTR030_USB_MAX_CONTROL_LEN;

        err = LTR_CrateCustomCtlReq(&module->Channel, LTR030_USB_CMD_WRITE_DSP_FLASH,
                                   addr, data, snd_size, NULL, 0, NULL);
        if (err==LTR_OK) {
            addr+=snd_size;
            size-=snd_size;
            data+=snd_size;
        }
    }
    return err;
}

LTR030API_DllExport(INT) LTR030_FlashRead(TLTR030 *module, DWORD addr, BYTE* data, DWORD size) {
    INT err = module == NULL ? LTR_ERROR_INVALID_MODULE_DESCR :
                               data == NULL ? LTR_ERROR_PARAMETERS : LTR_OK;
    while ((err==LTR_OK) && size) {
        DWORD rcv_size = size;
        if (rcv_size > LTR030_USB_MAX_CONTROL_LEN)
            rcv_size = LTR030_USB_MAX_CONTROL_LEN;

        err = LTR_CrateCustomCtlReq(&module->Channel, LTR030_USB_CMD_READ_DSP_FLASH,
                                   addr, NULL, 0, data, rcv_size, NULL);
        if (err==LTR_OK) {
            addr+=rcv_size;
            size-=rcv_size;
            data+=rcv_size;
        }
    }
    return err;
}

LTR030API_DllExport(INT) LTR030_SetInterfaceUsb(TLTR030 *module) {
    BYTE cfg[2];
    cfg[0] = LTR030_CONFIG_INTF_USB & 0xFF;
    cfg[1] = (LTR030_CONFIG_INTF_USB >> 8) & 0xFF;
    return LTR030_FlashWrite(module, LTR030_FLASH_ADDR_CONFIG, cfg, sizeof(cfg));
}




LTR030API_DllExport(INT) LTR030_SetInterfaceTcp(TLTR030 *module, DWORD ip_addr,
                                                DWORD ip_mask, DWORD gate) {
    BYTE cfg[LTR030_CONFIG_SIZE];
    cfg[0] = LTR030_CONFIG_INTF_ETH & 0xFF;
    cfg[1] = (LTR030_CONFIG_INTF_ETH >> 8) & 0xFF;
    cfg[2] = (ip_addr >> 24) & 0xFF;
    cfg[3] = (ip_addr >> 16) & 0xFF;
    cfg[4] = (ip_addr >>  8) & 0xFF;
    cfg[5] = ip_addr & 0xFF;
    cfg[6] = (gate >> 24) & 0xFF;
    cfg[7] = (gate >> 16) & 0xFF;
    cfg[8] = (gate >>  8) & 0xFF;
    cfg[9] = gate & 0xFF;
    cfg[10] = (ip_mask >> 24) & 0xFF;
    cfg[11] = (ip_mask >> 16) & 0xFF;
    cfg[12] = (ip_mask >>  8) & 0xFF;
    cfg[13] = ip_mask & 0xFF;
    return LTR030_FlashWrite(module, LTR030_FLASH_ADDR_CONFIG, cfg, sizeof(cfg));
}


LTR030API_DllExport(INT) LTR030_GetConfig(TLTR030 *module, TLTR030_CONFIG* cfg) {
    INT err = (cfg==NULL) || (cfg->Size < 2) ? LTR_ERROR_PARAMETERS : LTR_OK;
    if (err==LTR_OK) {
        BYTE raw_cfg[LTR030_CONFIG_SIZE];
        WORD ret_size = 2;
        err = LTR030_FlashRead(module, LTR030_FLASH_ADDR_CONFIG, raw_cfg, sizeof(raw_cfg));
        if (err==LTR_OK) {
            if (cfg->Size >= 4) {
                WORD intf_code = raw_cfg[0] | ((WORD)raw_cfg[1] << 8);
                cfg->Interface = intf_code == LTR030_CONFIG_INTF_ETH ? LTR_CRATE_IFACE_TCPIP :
                                 intf_code == LTR030_CONFIG_INTF_USB ? LTR_CRATE_IFACE_USB :
                                                                       LTR_CRATE_IFACE_UNKNOWN;
                ret_size = 4;
            }
            if (cfg->Size >= 8) {
                cfg->IpAddr = ((DWORD)raw_cfg[2] << 24) | ((DWORD)raw_cfg[3] << 16) |
                        ((DWORD)raw_cfg[4] << 8) | raw_cfg[5];
                ret_size = 8;
            }
            if (cfg->Size >= 12) {
                cfg->Gateway = ((DWORD)raw_cfg[6] << 24) | ((DWORD)raw_cfg[7] << 16) |
                        ((DWORD)raw_cfg[8] << 8) | raw_cfg[9];
                ret_size = 12;
            }
            if (cfg->Size >= 16) {
                cfg->IpMask = ((DWORD)raw_cfg[10] << 24) | ((DWORD)raw_cfg[11] << 16) |
                        ((DWORD)raw_cfg[12] << 8) | raw_cfg[13];
                ret_size = 16;
            }
            cfg->Size = ret_size;
        }
    }
    return err;
}



LTR030API_DllExport(INT) LTR030_SetConfig(TLTR030 *module, const TLTR030_CONFIG* cfg) {
    INT err = (cfg==NULL) || (cfg->Size < 4) ||
            ((cfg->Interface != LTR_CRATE_IFACE_TCPIP) && (cfg->Interface!=LTR_CRATE_IFACE_USB))
            ? LTR_ERROR_PARAMETERS : LTR_OK;

    if (err==LTR_OK) {
        BYTE raw_cfg[LTR030_CONFIG_SIZE];
        WORD wr_size = 0;
        WORD intf = cfg->Interface == LTR_CRATE_IFACE_TCPIP ? LTR030_CONFIG_INTF_ETH :
                                                              LTR030_CONFIG_INTF_USB;


        raw_cfg[wr_size++] = intf & 0xFF;
        raw_cfg[wr_size++] = (intf >> 8) & 0xFF;

        if (cfg->Size >= 8) {
            raw_cfg[wr_size++] = (cfg->IpAddr >> 24) & 0xFF;
            raw_cfg[wr_size++] = (cfg->IpAddr >> 16) & 0xFF;
            raw_cfg[wr_size++] = (cfg->IpAddr >>  8) & 0xFF;
            raw_cfg[wr_size++] = cfg->IpAddr & 0xFF;
        }
        if (cfg->Size >= 12) {
            raw_cfg[wr_size++] = (cfg->Gateway >> 24) & 0xFF;
            raw_cfg[wr_size++] = (cfg->Gateway  >> 16) & 0xFF;
            raw_cfg[wr_size++] = (cfg->Gateway  >>  8) & 0xFF;
            raw_cfg[wr_size++] = cfg->Gateway  & 0xFF;
        }
        if (cfg->Size >= 16) {
            raw_cfg[wr_size++] = (cfg->IpMask >> 24) & 0xFF;
            raw_cfg[wr_size++] = (cfg->IpMask >> 16) & 0xFF;
            raw_cfg[wr_size++] = (cfg->IpMask >>  8) & 0xFF;
            raw_cfg[wr_size++] = cfg->IpMask & 0xFF;
        }
        err = LTR030_FlashWrite(module, LTR030_FLASH_ADDR_CONFIG, raw_cfg, wr_size);
    }
    return err;
}



#define CALL_CB_LOAD()  if (cb) cb(stage, proc_size, total_size, cb_data);


static INT f_write_from_block(TLTR030* module, FILE* f, DWORD addr, DWORD max_size,
                              TLTR030_LOAD_PROGR_CB cb, void* cb_data) {
    INT err = LTR_OK;
    BYTE *block=NULL, *ver_block=NULL;
    DWORD total_size, proc_size=0;
    e_LTR030_LOAD_STAGES stage = LTR030_LOAD_STAGE_WRITE;

    block = malloc(LTR030_USB_MAX_CONTROL_LEN);
    ver_block = malloc(LTR030_USB_MAX_CONTROL_LEN);
    if ((block==NULL) || (ver_block==NULL)) {
        err = LTR_ERROR_MEMORY_ALLOC;
    } else {
        /* определяем полный разер файла */
        fseek(f, 0, SEEK_END);
        total_size = ftell(f);
        fseek(f, 0, SEEK_SET);

        if (total_size > max_size)
            err = LTR030_ERR_FIRM_SIZE;
    }

    if (err==LTR_OK) {
        CALL_CB_LOAD();

        while ((total_size-proc_size) && (err==LTR_OK)) {
            DWORD block_size = total_size-proc_size, rd_size;
            if (block_size>LTR030_USB_MAX_CONTROL_LEN)
                block_size = LTR030_USB_MAX_CONTROL_LEN;

            rd_size = (DWORD)fread(block, 1, block_size, f);
            if (rd_size!=block_size) {
                err = LTR_ERROR_FIRM_FILE_READ;
            } else {
                err = LTR030_FlashWrite(module, addr, block, block_size);
                if (err==LTR_OK) {
                    proc_size+=block_size;
                    addr+=block_size;
                    CALL_CB_LOAD();
                }
            }
        }

        if (err==LTR_OK) {
            stage = LTR030_LOAD_STAGE_VERIFY;
            proc_size = 0;
            addr-=total_size;
            fseek(f, 0, SEEK_SET);

            CALL_CB_LOAD();

            while ((total_size-proc_size) && (err==LTR_OK)) {
                DWORD block_size = total_size-proc_size, rd_size;
                if (block_size>LTR030_USB_MAX_CONTROL_LEN)
                    block_size = LTR030_USB_MAX_CONTROL_LEN;

                rd_size = (DWORD)fread(block, 1, block_size, f);
                if (rd_size!=block_size) {
                    err = LTR_ERROR_FIRM_FILE_READ;
                } else {
                    err = LTR030_FlashRead(module, addr, ver_block, block_size);
                    if (err==LTR_OK) {
                        unsigned i;
                        for (i=0; (i < block_size) && (err==LTR_OK); i++) {
                            if (block[i]!=ver_block[i])
                                err = LTR030_ERR_FIRM_VERIFY;
                        }

                    }
                    if (err==LTR_OK) {
                        proc_size+=block_size;
                        addr+=block_size;
                        CALL_CB_LOAD();
                    }
                }
            }
        }
    }

    free(block);
    free(ver_block);

    return err;
}



LTR030API_DllExport(INT) LTR030_LoadDspFirmware(TLTR030 *module, const char* filename,
                                                TLTR030_LOAD_PROGR_CB cb, void* cb_data) {
    INT err = module == NULL ? LTR_ERROR_INVALID_MODULE_DESCR : filename == NULL ?
                                   LTR_ERROR_PARAMETERS : LTR_OK;
    if (err==LTR_OK) {
        FILE *f = fopen(filename, "rb");

        if (f==NULL) {
            err = LTR_ERROR_FIRM_FILE_OPEN;
        } else {
            err = f_write_from_block(module, f, LTR030_FLASH_ADDR_DSP_FIRM,
                                    LTR030_FLASH_DSP_FIRM_MAX_SIZE,
                                     cb, cb_data);

            fclose(f);
        }
    }
    return err;
}


LTR030API_DllExport(INT) LTR030_LoadFpgaFirmware(TLTR030 *module, const char* filename,
                                                 TLTR030_LOAD_PROGR_CB cb, void* cb_data) {
    INT err = module == NULL ? LTR_ERROR_INVALID_MODULE_DESCR : filename == NULL ?
                                   LTR_ERROR_PARAMETERS : LTR_OK;
    if (err==LTR_OK) {
        FILE *f = fopen(filename, "rb");

        if (f==NULL) {
            err = LTR_ERROR_FIRM_FILE_OPEN;
        } else {
            BYTE *block = malloc(LTR030_USB_MAX_CONTROL_LEN);
            if (block==NULL) {
                err = LTR_ERROR_MEMORY_ALLOC;
            } else {
                BYTE info[LTR030_FPGA_INFO_SIZE];
                DWORD total_size, proc_size=0;
                DWORD chksum=0;

                fseek(f, 0, SEEK_END);
                total_size = ftell(f);
                fseek(f, 0, SEEK_SET);

                if (total_size > LTR030_FLASH_FPGA_FRIM_MAX_SIZE) {
                    err = LTR030_ERR_FIRM_SIZE;
                }

                while ((total_size-proc_size) && (err==LTR_OK)) {
                    DWORD block_size = total_size-proc_size, rd_size;
                    if (block_size>LTR030_USB_MAX_CONTROL_LEN)
                        block_size = LTR030_USB_MAX_CONTROL_LEN;

                    rd_size = (DWORD)fread(block, 1, block_size, f);
                    if (rd_size!=block_size) {
                        err = LTR_ERROR_FIRM_FILE_READ;
                    } else {
                        unsigned i;
                        for (i=0; i < block_size; i++) {
                            chksum = (chksum + block[i]) & 0xFFFFFFFFUL;
                        }

                        proc_size+=block_size;
                    }
                }

                if (err==LTR_OK) {
                    info[ 0 ] = total_size & 0xFF;
                    info[ 1 ] = (total_size >> 8) & 0xFF;
                    info[ 2 ] = (total_size >> 16) & 0xFF;
                    info[ 3 ] = (total_size >> 24) & 0xFF;
                    info[ 4 ] = chksum & 0xFF;
                    info[ 5 ] = (chksum >> 8) & 0xFF;
                    info[ 6 ] = (chksum >> 16) & 0xFF;
                    info[ 7 ] = (chksum >> 24) & 0xFF;
                    err = LTR030_FlashWrite(module, LTR030_FLASH_ADDR_FPGA_INFO,
                                        info, sizeof(info));
                }
                free(block);
            }

            if (err==LTR_OK) {
                err = f_write_from_block(module, f, LTR030_FLASH_ADDR_FPGA_FIRM,
                                         LTR030_FLASH_FPGA_FRIM_MAX_SIZE, cb, cb_data);
            }

            fclose(f);
        }
    }
    return err;
}


//-----------------------------------------------------------------------------
//
//-----------------------------------------------------------------------------
LTR030API_DllExport(LPCSTR)  LTR030_GetErrorString(INT error) {
    size_t i;
    for (i = 0; i < sizeof(f_err_str_tbl) / sizeof(f_err_str_tbl[0]); i++) {
        if (f_err_str_tbl[i].code == error)
            return f_err_str_tbl[i].message;
    }
    return LTR_GetErrorString(error);
}
//-----------------------------------------------------------------------------
#ifdef _WIN32
int WINAPI DllEntryPoint(HINSTANCE hinst, unsigned long reason, void* lpReserved) {
    return 1;
}
#endif

