#include <stddef.h>
#include <stdlib.h>
#include <string.h>

#include "ltr42api.h"
#include "ltr4x.h"
#include "ltrmodule.h"
#include "ltrslot.h"


/*================================================================================================*/
#pragma pack(4)
struct LTR42Config {
    CHAR name[16];
    CHAR serial[24];
    CHAR firmware_ver[8];
    CHAR firmware_date[16];
    BOOLEAN ack_en;
    struct {
        INT start;
        INT seconds;
    } marks_mode;
    WORD outputs_state;
};
#pragma pack()


/*================================================================================================*/
static void conv_hltr42_to_ltr42cfg(const void *h_card, void *cfg);
static void conv_ltr42cfg_to_hltr42(const void *cfg, void *h_card);
static INT LTR42_OpenEx(PTLTR42 h_ltr42, INT net_addr, WORD net_port, const CHAR *crate_sn,
    INT slot, DWORD in_flags, DWORD *out_flags);


/*================================================================================================*/
static const LPCSTR err_str[] = {
    "LTR42:Неверный указатель на структуру описания модуля!",
    "LTR42:Невозможно открыть модуль!",
    "LTR42:Неверный серийный номер крейта!",
    "LTR42:Неверный номер посадочного места!",
    "LTR42:Невозможно отправить команду!",
    "LTR42:Невозможно выполнить аппаратный сброс модуля!",
    "LTR42:Модуль не отвечает!",
    "LTR42:Невозможно отправить данные на выходные линии!",
    "LTR42:Невозможно выполнить конфигурацию модуля!",
    "LTR42:Невозможно запустить генерацию секундных меток!",
    "LTR42:Невозможно остановить генерацию секундных меток!",
    "LTR42:Невозможно сформировать метку СТАРТ!",
    "LTR42:Ошибка при передаче данных модулю!",
    "LTR42:Получено меньше слов, чем было заказано!",
    "LTR42:Ошибка четности при передаче команды модулю!",
    "LTR42:Ошибка четности при получении команды от модуля!",
    /* Посмотри, что у нас с СТАРТ и Сек! (0,1,2)? */
    "LTR42:Неверная конфигурация секундных меток! "
    "Только значения 0, 1 и 2 допустимы в Marks.SecondMark_Mode!",
    "LTR42:Неверная конфигурация метки СТАРТ! "
    "Только значения 0, 1 и 2 допустимы в Marks.StartMark_Mode!",
    "LTR42:Невозможно прочитать данные!",
    "LTR42:Невозможно выполнить запись в  EEPROM!",
    "LTR42:Невозможно выполнить чтение из EEPROM!",
    "LTR42:Неверный адрес EEPROM!",
    "LTR42:Невозможно считать идентификационную запись!",
    "LTR42:Неверная идентификационная запись!"
};
static const int min_range_id = -8000;


/*================================================================================================*/
/*------------------------------------------------------------------------------------------------*/
static void conv_hltr42_to_ltr42cfg(const void *h_card, void *cfg) {
    const TLTR42 *h_ltr42 = h_card;
    struct LTR42Config *pcfg = cfg;

    strncpy(pcfg->name, h_ltr42->ModuleInfo.Name, sizeof pcfg->name / sizeof pcfg->name[0]);
    strncpy(pcfg->serial, h_ltr42->ModuleInfo.Serial, sizeof pcfg->serial / sizeof pcfg->serial[0]);
    strncpy(pcfg->firmware_ver, h_ltr42->ModuleInfo.FirmwareVersion,
        sizeof pcfg->firmware_ver / sizeof pcfg->firmware_ver[0]);
    strncpy(pcfg->firmware_date, h_ltr42->ModuleInfo.FirmwareDate,
        sizeof pcfg->firmware_date / sizeof pcfg->firmware_date[0]);

    pcfg->ack_en = h_ltr42->AckEna;
    pcfg->marks_mode.start = h_ltr42->Marks.StartMark_Mode;
    pcfg->marks_mode.seconds = h_ltr42->Marks.SecondMark_Mode;
}

/*------------------------------------------------------------------------------------------------*/
static void conv_ltr42cfg_to_hltr42(const void *cfg, void *h_card) {
    PTLTR42 h_ltr42 = h_card;
    const struct LTR42Config *pcfg = cfg;

    strncpy(h_ltr42->ModuleInfo.Name, pcfg->name,
        sizeof h_ltr42->ModuleInfo.Name / sizeof h_ltr42->ModuleInfo.Name[0]);
    strncpy(h_ltr42->ModuleInfo.Serial, pcfg->serial,
        sizeof h_ltr42->ModuleInfo.Serial / sizeof h_ltr42->ModuleInfo.Serial[0]);
    strncpy(h_ltr42->ModuleInfo.FirmwareVersion, pcfg->firmware_ver,
        sizeof h_ltr42->ModuleInfo.FirmwareVersion / sizeof h_ltr42->ModuleInfo.FirmwareVersion[0]);
    strncpy(h_ltr42->ModuleInfo.FirmwareDate, pcfg->firmware_date,
        sizeof h_ltr42->ModuleInfo.FirmwareDate / sizeof h_ltr42->ModuleInfo.FirmwareDate[0]);

    h_ltr42->AckEna = pcfg->ack_en;
    h_ltr42->Marks.StartMark_Mode = pcfg->marks_mode.start;
    h_ltr42->Marks.SecondMark_Mode = pcfg->marks_mode.seconds;
}

/*------------------------------------------------------------------------------------------------*/
#ifdef _WIN32
int __stdcall DllEntryPoint(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved) {
    /* Included for compatibility with Borland */
    return 1;
}
#endif

/*------------------------------------------------------------------------------------------------*/
#ifdef _WIN32
BOOL APIENTRY DllMain(HMODULE hModule, DWORD  ul_reason_for_call, LPVOID lpReserved) {
    return TRUE;
}
#endif

/*------------------------------------------------------------------------------------------------*/
LTR42API_DllExport(INT) LTR42_Close(PTLTR42 hnd) {
    /* Функция окончания работы с модулем */
    INT err = (hnd == NULL) ? LTR_ERROR_INVALID_MODULE_DESCR : LTR_OK;
    if (err == LTR_OK) {
        INT stop_err;
        err = LTR42_StopSecondMark(hnd);
        stop_err = LTR_Close(&hnd->Channel);
        if (err == LTR_OK)
            err = stop_err;
    }
    return err;
}

/*------------------------------------------------------------------------------------------------*/
LTR42API_DllExport(INT) LTR42_Config(PTLTR42 hnd) {
    /* Функция конфигурирования модуля */
    INT err = (hnd == NULL) ? LTR_ERROR_INVALID_MODULE_DESCR : LTR_OK;
    if (err == LTR_OK) {
        if((hnd->Marks.StartMark_Mode < 0) || (hnd->Marks.StartMark_Mode > 2)) {
            err = LTR42_ERR_WRONG_START_MARK_CONF;
        } else if ((hnd->Marks.SecondMark_Mode < 0) || (hnd->Marks.StartMark_Mode > 2)) {
            err = LTR42_ERR_WRONG_SECOND_MARK_CONF;
        }
    }

    if (err == LTR_OK) {
        DWORD command;
        DWORD command_ack;
        if (hnd->AckEna != 0)
            hnd->AckEna = 1;

        command = ltr_module_fill_cmd_parity(CONFIG, (((WORD)hnd->AckEna) << 8) |
            (hnd->Marks.StartMark_Mode << 4) | hnd->Marks.SecondMark_Mode);
        err = LTR4X_SendCmdWithAck(&hnd->Channel, &command, &command_ack, 1);
    }


    return err;
}

/*------------------------------------------------------------------------------------------------*/
LTR42API_DllExport(INT) LTR42_ConfigAndStart(PTLTR42 hnd) {
    /* Функция конфигурирования модуля */
    INT err = (hnd == NULL) ? LTR_ERROR_INVALID_MODULE_DESCR : LTR_OK;
    if (err == LTR_OK) {
        if((hnd->Marks.StartMark_Mode < 0) || (hnd->Marks.StartMark_Mode > 2)) {
            err = LTR42_ERR_WRONG_START_MARK_CONF;
        } else if ((hnd->Marks.SecondMark_Mode < 0) || (hnd->Marks.StartMark_Mode > 2)) {
            err = LTR42_ERR_WRONG_SECOND_MARK_CONF;
        }
    }

    if (err == LTR_OK) {
        DWORD command;
        DWORD command_ack;
        if (hnd->AckEna != 0)
            hnd->AckEna = 1;

        command = ltr_module_fill_cmd_parity(CONFIG, (((WORD)hnd->AckEna) << 8) |
            (hnd->Marks.StartMark_Mode << 4) | hnd->Marks.SecondMark_Mode);
        err = LTR4X_SendCmdWithAck(&hnd->Channel, &command, &command_ack, 1);
    }

    if (err == LTR_OK) {
        err = ltrslot_start_wconfig(hnd, &hnd->Channel, sizeof(struct LTR42Config), LTR_MID_LTR42,
            conv_hltr42_to_ltr42cfg);
    }

    return err;
}

/*------------------------------------------------------------------------------------------------*/
LTR42API_DllExport(LPCSTR) LTR42_GetErrorString(int Error_Code) {
    /* Функция определения строки ошибки */
    INT error_range;
    int ret_val_index;
    LPCSTR ret_val = NULL;

    error_range = -min_range_id + 1 + sizeof err_str / sizeof err_str[0];

    if ((-Error_Code < -min_range_id ) || (Error_Code < -error_range)) {
        ret_val = LTR_GetErrorString(Error_Code);
    } else {
        ret_val_index = -Error_Code + min_range_id - 1;
        ret_val = err_str[ret_val_index];
    }

    return ret_val;
}

/*------------------------------------------------------------------------------------------------*/
LTR42API_DllExport(INT) LTR42_Init(PTLTR42 hnd) {
    /* Функция заполнения структуры описания модуля значениями по умолчанию и
     * инициализации интерфейсного канала
     */
    INT err = (hnd == NULL) ? LTR_ERROR_INVALID_MODULE_DESCR : LTR_OK;
    if (err == LTR_OK) {
        hnd->size = sizeof(TLTR42);

        strcpy(hnd->ModuleInfo.FirmwareVersion,"\0");
        strcpy(hnd->ModuleInfo.FirmwareDate,"\0");
        strcpy(hnd->ModuleInfo.Serial,"\0");
        strcpy(hnd->ModuleInfo.Name,"\0");

        hnd->AckEna = 1;                    /* По умолчанию подтверждения включены */
        err = LTR_Init(&hnd->Channel);      /* Инициализируем интерефейсный канал */
    }
    return err;
}

/*------------------------------------------------------------------------------------------------*/
LTR42API_DllExport(INT) LTR42_IsOpened(PTLTR42 hnd) {
    /* Функция проверки: открыт ли модуль */
    return (hnd == NULL) ? LTR_ERROR_INVALID_MODULE_DESCR : LTR_IsOpened(&hnd->Channel);
}

/*------------------------------------------------------------------------------------------------*/
LTR42API_DllExport(INT) LTR42_MakeStartMark(PTLTR42 hnd) {
    /* Функция формирования метки "СТАРТ" */
    INT err = (hnd == NULL) ? LTR_ERROR_INVALID_MODULE_DESCR : LTR_OK;
    if (err == LTR_OK) {
        DWORD ack;
        DWORD command;
        command = ltr_module_fill_cmd_parity(MAKE_START_MARK, 0);
        err = LTR4X_SendCmdWithAck(&hnd->Channel, &command, &ack, 1);
    }
    return err;
}

/*------------------------------------------------------------------------------------------------*/
LTR42API_DllExport(INT) LTR42_Open(PTLTR42 hnd, INT net_addr, WORD net_port, const CHAR *crate_sn,
    INT slot_num) {
    /* Функция создания и открытия канала связи с LTR42. Также RESET модуля */
    return LTR42_OpenEx(hnd, net_addr, net_port, crate_sn, slot_num, 0, NULL);
}

/*------------------------------------------------------------------------------------------------*/
static INT LTR42_OpenEx(PTLTR42 h_ltr42, INT net_addr, WORD net_port, const CHAR *crate_sn,
    INT slot, DWORD in_flags, DWORD *out_flags) {
    /* Функция создания и открытия канала связи с LTR42. Также RESET модуля */

    DWORD open_flg = 0;
    DWORD out_flg = 0;
    INT ret_val = (h_ltr42 == NULL) ? LTR_ERROR_INVALID_MODULE_DESCR : LTR_OK;

    if (ret_val == LTR_OK) {
        if (in_flags & LTR_OPENINFLG_REOPEN) {
            ret_val = ltrslot_restore_config(h_ltr42, net_addr, net_port, crate_sn, slot,
                sizeof(struct LTR42Config), conv_ltr42cfg_to_hltr42, &out_flg);
            if (ret_val == LTR_OK) {
                if (out_flg & LTR_OPENOUTFLG_REOPEN)
                    open_flg |= LTR_MOPEN_INFLAGS_DONT_RESET;
            }
        }
    }

    if (ret_val == LTR_OK) {
        ret_val = LTR4X_OpenEx(&h_ltr42->Channel, net_addr, net_port, crate_sn, slot, open_flg,
            &h_ltr42->ModuleInfo, NULL);
    }

    if (out_flags != NULL)
        *out_flags = out_flg;

    return ret_val;
}

/*------------------------------------------------------------------------------------------------*/
LTR42API_DllExport(INT) LTR42_ReadEEPROM(PTLTR42 hnd, INT Address, BYTE *val) {
    /* Функция чтения байта из указанной ячейки ППЗУ */
    INT err = (hnd == NULL) ? LTR_ERROR_INVALID_MODULE_DESCR : LTR_OK;
    if (err == LTR_OK)
        err = LTR4X_ReadEEPROM(&hnd->Channel, Address, val);
    return err;
}

/*------------------------------------------------------------------------------------------------*/
LTR42API_DllExport(INT) LTR42_StartSecondMark(PTLTR42 hnd) {
    /* Функция включения генерации секундной метки */
    INT err = (hnd == NULL) ? LTR_ERROR_INVALID_MODULE_DESCR : LTR_OK;
    if (err == LTR_OK) {
        DWORD ack;
        DWORD command;
        command = ltr_module_fill_cmd_parity(START_SECOND_MARK, 0);
        err = LTR4X_SendCmdWithAck(&hnd->Channel, &command, &ack, 1);
    }
    return err;
}

/*------------------------------------------------------------------------------------------------*/
LTR42API_DllExport(INT) LTR42_StopSecondMark(PTLTR42 hnd) {
    /* Функция выключения генерации внутренней секундной метки */
    INT err = (hnd == NULL) ? LTR_ERROR_INVALID_MODULE_DESCR : LTR_OK;
    if (err == LTR_OK) {
        DWORD ack;
        DWORD command;
        command = ltr_module_fill_cmd_parity(STOP_SECOND_MARK, 0);
        err = LTR4X_SendCmdWithAck(&hnd->Channel, &command, &ack, 1);
    }
    return err;
}

/*------------------------------------------------------------------------------------------------*/
LTR42API_DllExport(INT) LTR42_StoreConfig(PTLTR42 hnd, TLTR_CARD_START_MODE start_mode) {
    INT ret_code = LTR42_IsOpened(hnd);

    if (ret_code == LTR_OK) {
        ret_code = ltrslot_store_config(hnd, &hnd->Channel, sizeof(struct LTR42Config), LTR_MID_LTR42,
            start_mode, conv_hltr42_to_ltr42cfg);
    }

    return ret_code;
}

/*------------------------------------------------------------------------------------------------*/
LTR42API_DllExport(INT) LTR42_WriteArray(PTLTR42 hnd, const WORD *OutputArray, INT ArraySize) {
    /* Функция отправки массива 16-битных слов в выходной порт модуля */
    INT i;
    INT err = (hnd == NULL) ? LTR_ERROR_INVALID_MODULE_DESCR :
        ((OutputArray == NULL) || (ArraySize <= 0)) ? LTR_ERROR_PARAMETERS : LTR_OK;
    for (i = 0; ((i < ArraySize) && (err == LTR_OK)); i++) {
        DWORD data[2];
        data[0] = (*OutputArray << 16) | 1; /* Отправляем пакет со словом для выхода */
        data[1] = (*OutputArray << 16);     /* Дублируем пакет */
        OutputArray++;
        /* @todo здесь был некий счетчик частоты передачи данных, который никак не использовался
         * (мб кто то в отладчике смотрел) и закоментированный Sleep(). Вообще по идее тут надо
         * как-то умнее действовать, чем просто тупо слать сперва все, а потом принимать все ответы,
         * т.к. иначе LTR42 может не успеть
         */
        err = ltr_module_send_cmd(&hnd->Channel, data, sizeof(data)/sizeof(data[0]));
    }

    if ((err == LTR_OK) && hnd->AckEna) {
        for (i = 0; ((i < ArraySize) && (err == LTR_OK)); i++) {
            DWORD command_ack;
            err = ltr_module_recv_cmd_resp(&hnd->Channel, &command_ack, 1);
            if (err == LTR_OK)
                err = LTR4X_CheckAck(command_ack, DATA_OUTPUT_CONFIRM);
        }
    }
    return err;
}

/*------------------------------------------------------------------------------------------------*/
LTR42API_DllExport(INT) LTR42_WriteEEPROM(PTLTR42 hnd, INT Address, BYTE val) {
    /* Функция записи байта в указанную ячейку ППЗУ */
    INT err = (hnd == NULL) ? LTR_ERROR_INVALID_MODULE_DESCR : LTR_OK;
    if (err == LTR_OK)
        err = LTR4X_WriteEEPROM(&hnd->Channel, Address, val);
    return err;
}

/*------------------------------------------------------------------------------------------------*/
LTR42API_DllExport(INT) LTR42_WritePort(PTLTR42 hnd, WORD OutputData) {
    /* Функция отправки 16-битного слова в выходной порт модуля */
    INT err = (hnd == NULL) ? LTR_ERROR_INVALID_MODULE_DESCR : LTR_OK;
    if (err == LTR_OK) {
        DWORD data[2];
        data[0] = (OutputData << 16) | 1;   /* Отправляем пакет со словом для выхода */
        data[1] = (OutputData << 16);       /* Дублируем пакет */

        err = ltr_module_send_cmd(&hnd->Channel, data, sizeof(data)/sizeof(data[0]));
        /* Если включены подтверждения записи слов - принимаем и проверяем подтверждение */
        if ((err == LTR_OK) && hnd->AckEna) {
            DWORD command_ack;
            err = ltr_module_recv_cmd_resp(&hnd->Channel, &command_ack, 1);
            if (err == LTR_OK)
                err = LTR4X_CheckAck(command_ack, DATA_OUTPUT_CONFIRM);
        }
    }
    return err;
}

/*------------------------------------------------------------------------------------------------*/
LTR42API_DllExport(INT) LTR42_WritePortSaved(PTLTR42 hnd, WORD OutputData) {
    /* Функция отправки 16-битного слова в выходной порт модуля с сохранением его состояния в
     * крейте
     */
    int slots_supported;
    t_ltrslot_conn h_slot_conn = LTRSLOT_CONN_INVAL;
    INT err = (hnd == NULL) ? LTR_ERROR_INVALID_MODULE_DESCR : LTR_OK;
    if (err == LTR_OK) {
        DWORD data[2];
        data[0] = (OutputData << 16) | 1;   /* Отправляем пакет со словом для выхода */
        data[1] = (OutputData << 16);       /* Дублируем пакет */

        err = ltr_module_send_cmd(&hnd->Channel, data, sizeof(data)/sizeof(data[0]));
        /* Если включены подтверждения записи слов - принимаем и проверяем подтверждение */
        if ((err == LTR_OK) && hnd->AckEna) {
            DWORD command_ack;
            err = ltr_module_recv_cmd_resp(&hnd->Channel, &command_ack, 1);
            if (err == LTR_OK)
                err = LTR4X_CheckAck(command_ack, DATA_OUTPUT_CONFIRM);
        }
    }

    if (err == LTR_OK)
        err = ltrslot_replicate_connection(&hnd->Channel, &h_slot_conn);

    if (err == LTR_OK)
        err = ltrslot_check_support(h_slot_conn, &slots_supported);

    if (err == LTR_OK) {
        if (!slots_supported)
            err = LTR_ERROR_CARDSCONFIG_UNSUPPORTED;
    }

    if (err == LTR_OK) {
        const WORD outs = OutputData;
        err = ltrslot_update_config(h_slot_conn, offsetof(struct LTR42Config, outputs_state),
            sizeof(WORD), &outs);
    }


    ltrslot_destroy_connection(h_slot_conn);

    return err;
}
